import numpy as np
from subprocess import call
import time
import pkg_resources
from sarabande.utils.functions import *


def calc_zeta_mapped(
    measure_obj, normalize=True, verbose_flag=True, skip_prepare=False
):
    """
    This function is where the core algorithms take place for measuring the 3/4 PCFs 
    either projected or not projected. In total there are 4 options

    - projected 3PCF        - projected 4PCF
    - full 3PCF             - full 4PCF
    
    Args:
        measure_obj (class: measure): an object that carries the necessary values and data structures to compute 3/4 PCFs
        normalize (bool, optional): flag to determine whether we would like to normalize by the bin volume(s). Defaults to True.
        verbose_flag (bool, optional): flag to walk the user through the full calculation process. Defaults to True.
        skip_prepare (bool, optional): Flag to determine whether or not we should skip preparing the data (if you already ran the calculation). Defaults to False.

    Raises:
        AssertionError: in order to calculate the full 4PCF one has to save a boundsandnumber file (this comes from creating radial bins in one of the utility functions.)

    Returns:
        No Returns, instead we add the zeta attribute to the measure_obj.
    """

    ##############################################################
    #                Projected Zeta Calculation
    ##############################################################
    if measure_obj.projected:
        start = time.time()
        if not skip_prepare:
            print("""Preparing the data:""")
            projected_prepare_data(measure_obj, verbose_flag)
        else:
            # fix bounds and number scope issue
            measure_obj.kernel_name = measure_obj.save_name
            # measure_obj.boundsandnumber = np.load(measure_obj.save_dir + 'bin_bounds_and_pixel_number_'+measure_obj.save_name+'.npy')

        if measure_obj.nPCF == 3:
            if verbose_flag:
                print("Starting Calculation of Projected 3PCF.")

            PCF_computed = (
                np.zeros((measure_obj.nbins, measure_obj.nbins, measure_obj.m_max)) + 0j
            )
            for bin1 in range(measure_obj.nbins):
                for bin2 in range(bin1):
                    for m in range(measure_obj.m_max):
                        PCF_computed[bin1, bin2, m] += np.sum(
                            measure_obj.density_field_data
                            * (
                                measure_obj.conv_M[m, bin1, :, :]
                                * measure_obj.conv_M[m, bin2, :, :].conjugate()
                            )
                        )

            PCF_computed_ = PCF_computed.copy()
            if normalize:
                # Normalize Full
                for bin1 in range(measure_obj.nbins):
                    for bin2 in range(bin1 + 1):
                        for m in range(measure_obj.m_max):
                            PCF_computed_[bin1, bin2, m] /= (
                                measure_obj.boundsandnumber[1, bin1]
                                * measure_obj.boundsandnumber[1, bin2]
                            )
                            PCF_computed_[bin2, bin1, m] = PCF_computed_[bin1, bin2, m]

                measure_obj.zeta = PCF_computed_
            else:
                measure_obj.zeta = PCF_computed_

            stop = time.time()

            print(
                "\nFinished Calculating the Projected 3PCF in {0:0.4f} seconds".format(
                    stop - start
                )
            )
            # coeff_1d = PCF_computed_1d_

        elif measure_obj.nPCF == 4:
            if verbose_flag:
                print("Starting Calculation of Projected 4PCF.")

            # compute the 4PCF
            PCF_computed = (
                np.zeros(
                    (
                        measure_obj.nbins,
                        measure_obj.nbins,
                        measure_obj.nbins,
                        2 * measure_obj.m_max,
                        2 * measure_obj.m_max,
                    )
                )
                + 0j
            )
            for bin1 in range(measure_obj.nbins):
                for bin2 in range(bin1):
                    for bin3 in range(bin2):
                        for m1 in range(-measure_obj.m_max, measure_obj.m_max):
                            if m1 > 0:
                                c_m1 = measure_obj.conv_M[m1, bin1, :, :]
                            else:
                                c_m1 = (-1) ** m1 * measure_obj.conv_M[
                                    -m1, bin1, :, :
                                ].conjugate()

                            for m2 in range(-measure_obj.m_max, measure_obj.m_max):
                                if m2 > 0:
                                    c_m2 = measure_obj.conv_M[m2, bin2, :, :]
                                else:
                                    c_m2 = (-1) ** m2 * measure_obj.conv_M[
                                        -m2, bin2, :, :
                                    ].conjugate()

                                m3 = -(m1 + m2)
                                if abs(m3) >= measure_obj.m_max:
                                    continue
                                else:
                                    if m3 > 0:
                                        c_m3 = measure_obj.conv_M[m3, bin3, :, :]
                                    else:
                                        c_m3 = (-1) ** m3 * measure_obj.conv_M[
                                            -m3, bin3, :, :
                                        ].conjugate()

                                    PCF_computed[bin1, bin2, bin3, m1, m2] += np.sum(
                                        measure_obj.density_field_data
                                        * (c_m1 * c_m2 * c_m3)
                                    )

            PCF_computed_ = PCF_computed.copy()
            if normalize:
                # Normalize and Fill
                for bin1 in range(measure_obj.nbins):
                    for bin2 in range(bin1):
                        for bin3 in range(bin2):
                            for m1 in range(-measure_obj.m_max, measure_obj.m_max):
                                for m2 in range(-measure_obj.m_max, measure_obj.m_max):
                                    m3 = -(m1 + m2)
                                    if abs(m3) >= measure_obj.m_max:
                                        continue
                                    else:
                                        PCF_computed_[bin1, bin2, bin3, m1, m2] /= (
                                            measure_obj.boundsandnumber[1, bin1]
                                            * measure_obj.boundsandnumber[1, bin2]
                                            * measure_obj.boundsandnumber[1, bin3]
                                        )
                                        PCF_computed_[
                                            bin3, bin1, bin2, m1, m2
                                        ] = PCF_computed_[bin1, bin2, bin3, m1, m2]
                                        PCF_computed_[
                                            bin2, bin3, bin1, m1, m2
                                        ] = PCF_computed_[bin1, bin2, bin3, m1, m2]
                                        PCF_computed_[
                                            bin1, bin3, bin2, m1, m2
                                        ] = PCF_computed_[bin1, bin2, bin3, m1, m2]
                                        PCF_computed_[
                                            bin2, bin1, bin3, m1, m2
                                        ] = PCF_computed_[bin1, bin2, bin3, m1, m2]
                                        PCF_computed_[
                                            bin3, bin2, bin1, m1, m2
                                        ] = PCF_computed_[bin1, bin2, bin3, m1, m2]

                measure_obj.zeta = PCF_computed_
            else:
                measure_obj.zeta = PCF_computed_

            stop = time.time()

            print(
                "\nFinished Calculating the Projected 4PCF in {0:0.4f} seconds".format(
                    stop - start
                )
            )

    ##############################################################
    #                   Full Zeta Calculation
    ##############################################################
    else:

        if not skip_prepare:
            print("""Preparing the data:""")
            prepare_data(measure_obj, verbose_flag)
        else:
            measure_obj.kernel_name = measure_obj.save_name
            # if not hasattr(measure_obj, 'boundsandnumber'):
            #     measure_obj.boundsandnumber = np.load(measure_obj.save_dir + 'bin_bounds_and_pixel_number_'+measure_obj.save_name+'.npy')

        if measure_obj.nPCF == 3:
            start = time.time()

            # load alms (results of convolution at fixed ell and for all m, to combine).
            zeta = (
                np.zeros(
                    (measure_obj.ell_max + 1, measure_obj.nbins, measure_obj.nbins)
                )
                + 0j
            )

            for l in range(0, measure_obj.ell_max + 1, 1):
                for bin1 in range(0, measure_obj.nbins, 1):
                    for bin2 in range(0, bin1, 1):
                        for m in range(0, l + 1, 1):
                            # load b1 and b2 at this m

                            ylm_b1 = np.load(
                                measure_obj.save_dir
                                + measure_obj.save_name
                                + "conv_data_kernel_"
                                + measure_obj.kernel_name
                                + "_"
                                + str(l)
                                + "_"
                                + str(m)
                                + "_bin_"
                                + str(bin1)
                                + ".npy"
                            ).astype(np.complex128)
                            ylm_b2 = np.load(
                                measure_obj.save_dir
                                + measure_obj.save_name
                                + "conv_data_kernel_"
                                + measure_obj.kernel_name
                                + "_"
                                + str(l)
                                + "_"
                                + str(m)
                                + "_bin_"
                                + str(bin2)
                                + ".npy"
                            ).astype(np.complex128)
                            # form half of sum at that m.
                            ylm_b1 *= ylm_b2.conjugate()
                            del ylm_b2

                            # average spatially now.
                            ylm_avg = np.sum(measure_obj.density_field_data * ylm_b1)
                            del ylm_b1
                            # add in negative m: so basically we are computing m's in sum term by
                            # term but with matching between + and - m.
                            if m > 0:
                                ylm_avg += ylm_avg.conjugate()
                            zeta[
                                l, bin1, bin2
                            ] += ylm_avg  # now accumulate so we get sum over m.

            zeta *= 1.0 / (4 * np.pi)

            for l in range(0, measure_obj.ell_max + 1, 1):
                for bin1 in range(0, measure_obj.nbins, 1):
                    for bin2 in range(
                        bin1, measure_obj.nbins, 1
                    ):  # ?? for bin2 in range(0, bin2<=bin1, 1):
                        zeta[l, bin1, bin2] = zeta[l, bin2, bin1]

            if normalize == True:
                binvolume = measure_obj.boundsandnumber[1, 0 : measure_obj.nbins]
                normedzeta = zeta / (binvolume[:, None] * binvolume[None, :])
                measure_obj.zeta = normedzeta

            else:
                measure_obj.zeta = zeta

            end = time.time()
            print("\n3PCF took {0:0.4f} seconds to finish".format(end - start))

            call(
                "rm "
                + measure_obj.save_dir
                + "bin_bounds_and_pixel_number_"
                + measure_obj.save_name
                + "*",
                shell=True,
            )
            call(
                "rm "
                + measure_obj.save_dir
                + measure_obj.save_name
                + "conv_data_kernel_"
                + measure_obj.kernel_name
                + "*",
                shell=True,
            )

        elif measure_obj.nPCF == 4:

            """
            This is the the big calculation for measuring the 4PCF coefficients.
            This code is adapted from Philcox et al. 2021 (Encore Paper)
            """

            def S(m):
                """
                A simple function defined in section 4.1.1 of
                Philcox et al. 2021 above algorithm 3. (Helper Function)
                """
                if m == 0:
                    return 1 / 2
                return 1

            def complex_modulus(z):
                """
                assume z to be an array of complex numbers of the form z = a + ib
                (Helper Function)
                """
                a = np.real(z)
                b = np.imag(z)

                return np.sqrt(a ** 2 + b ** 2)

            def load_almb(index_list):
                l_1, l_2, l_3, m_1, m_2, m_3, b_1, b_2, b_3 = index_list

                coupling_w = (
                    measure_obj.density_field_data
                    * (-1) ** (l_1 + l_2 + l_3)
                    * CG_Coefficients[l_1, l_2, l_3, m_1, m_2, m_3]
                )

                if m_1 < 0:
                    a_lmb_1 = (-1) ** m_1 * (
                        np.load(
                            measure_obj.save_dir
                            + measure_obj.save_name
                            + "conv_data_kernel_"
                            + measure_obj.kernel_name
                            + "_"
                            + str(l_1)
                            + "_"
                            + str(-m_1)
                            + "_bin_"
                            + str(b_1)
                            + ".npy"
                        ).astype(np.complex128)
                    ).conjugate()
                else:
                    a_lmb_1 = np.load(
                        measure_obj.save_dir
                        + measure_obj.save_name
                        + "conv_data_kernel_"
                        + measure_obj.kernel_name
                        + "_"
                        + str(l_1)
                        + "_"
                        + str(m_1)
                        + "_bin_"
                        + str(b_1)
                        + ".npy"
                    ).astype(np.complex128)

                if m_2 < 0:
                    a_lmb_2 = (-1) ** m_2 * (
                        np.load(
                            measure_obj.save_dir
                            + measure_obj.save_name
                            + "conv_data_kernel_"
                            + measure_obj.kernel_name
                            + "_"
                            + str(l_2)
                            + "_"
                            + str(-m_2)
                            + "_bin_"
                            + str(b_2)
                            + ".npy"
                        ).astype(np.complex128)
                    ).conjugate()
                else:
                    a_lmb_2 = np.load(
                        measure_obj.save_dir
                        + measure_obj.save_name
                        + "conv_data_kernel_"
                        + measure_obj.kernel_name
                        + "_"
                        + str(l_2)
                        + "_"
                        + str(m_2)
                        + "_bin_"
                        + str(b_2)
                        + ".npy"
                    ).astype(np.complex128)

                a_lmb_3 = np.load(
                    measure_obj.save_dir
                    + measure_obj.save_name
                    + "conv_data_kernel_"
                    + measure_obj.kernel_name
                    + "_"
                    + str(l_3)
                    + "_"
                    + str(m_3)
                    + "_bin_"
                    + str(b_3)
                    + ".npy"
                ).astype(np.complex128)

                zeta[l_1, l_2, l_3, b_1, b_2, b_3] += np.sum(
                    2 * S(m_3) * coupling_w * np.real(a_lmb_1 * a_lmb_2 * a_lmb_3)
                )

            # we define these local variables out of laziness to change my original non-object oriented code
            ell_max = measure_obj.ell_max
            nbins = measure_obj.nbins
            if hasattr(measure_obj, "boundsandnumber") == False:
                raise AssertionError(
                    "You need to run measure_obj.create_radial_bins() first!"
                )

            binvolume = measure_obj.boundsandnumber[1, 0:nbins]
            zeta = (
                np.zeros((ell_max + 1, ell_max + 1, ell_max + 1, nbins, nbins, nbins))
                + 0j
            )
            stream = pkg_resources.resource_stream(__name__, "utils/CG_Coeffs.npy")
            CG_Coefficients = np.load(stream)

            indeces = []
            for l_1 in range(0, ell_max + 1):
                for l_2 in range(0, ell_max + 1):
                    for l_3 in range(np.abs(l_1 - l_2), min(l_1 + l_2, ell_max) + 1):
                        if (
                            l_1 + l_2 + l_3
                        ) % 2 != 0:  # we don't assume this to be true for Turbulent ISM
                            continue
                        for m_1 in range(-l_1, l_1 + 1):
                            for m_2 in range(-l_2, l_2 + 1):
                                m_3 = -m_1 - m_2
                                if m_3 > l_3 or m_3 < 0:
                                    continue
                                for b_1 in range(0, nbins):
                                    for b_2 in range(b_1 + 1, nbins):
                                        for b_3 in range(b_2 + 1, nbins):
                                            indeces.append(
                                                (
                                                    l_1,
                                                    l_2,
                                                    l_3,
                                                    m_1,
                                                    m_2,
                                                    m_3,
                                                    b_1,
                                                    b_2,
                                                    b_3,
                                                )
                                            )

            start = time.time()
            print("Executing 4PCF Calculation ...")
            list(map(load_almb, indeces))  # speed up over original for loop structure

            for l1 in range(0, ell_max + 1):
                for l2 in range(0, ell_max + 1):
                    for l3 in range(np.abs(l1 - l2), min(l1 + l2, ell_max) + 1):
                        if (
                            l1 + l2 + l3
                        ) % 2 != 0:  # we don't assume this to be true for Turbulent ISM
                            continue
                        for b1 in range(0, nbins):
                            for b2 in range(b1 + 1, nbins):
                                for b3 in range(b2 + 1, nbins):
                                    this_4pcf = zeta[l1, l2, l3, b1, b2, b3]
                                    zeta[l3, l1, l2, b3, b1, b2] = this_4pcf
                                    zeta[l2, l3, l1, b2, b3, b1] = this_4pcf
                                    zeta[l1, l3, l2, b1, b3, b2] = this_4pcf
                                    zeta[l2, l1, l3, b2, b1, b3] = this_4pcf
                                    zeta[l3, l2, l1, b3, b2, b1] = this_4pcf

            finish = time.time()

            call(
                "rm "
                + measure_obj.save_dir
                + "bin_bounds_and_pixel_number_"
                + measure_obj.save_name
                + "*",
                shell=True,
            )
            call(
                "rm "
                + measure_obj.save_dir
                + measure_obj.save_name
                + "conv_data_kernel_"
                + measure_obj.kernel_name
                + "*",
                shell=True,
            )

            print(
                "Finished Calculating 4PCF in {0:0.4f} seconds".format(finish - start)
            )
            if normalize:
                """
                Normalize zeta^L_B (where L = {\ell_1, \ell_2, \ell_3} and B = {b_1, b_2, b_3}) hat
                coefficients from calc_zeta by dividing by bin volume
                """
                # binvolume = measure_obj.boundsandnumber[1,0:nbins]
                # normalize_coeff = (binvolume[:,None, None] * binvolume[None,:, None] * binvolume[None, None, :])
                normalize_coeff = (4.0 * np.pi) ** (3.0)
                zeta = normalize_coeff * zeta / ((measure_obj.ld_one_d ** 3))
                measure_obj.zeta = zeta
            else:
                print("your zeta coefficients are not properly ormalized.")
                measure_obj.zeta = zeta



import numpy as np
from subprocess import call
import time
import pkg_resources
import concurrent.futures
from sarabande.utils.functions import *

def calc_zeta_parallel(measure_obj, normalize=True, verbose_flag=True, skip_prepare=False):
    """
    This function is where the core algorithms take place for measuring the 3/4 PCFs 
    either projected or not projected. In total there are 4 options

    - projected 3PCF        - projected 4PCF
    - full 3PCF             - full 4PCF
    
    Args:
        measure_obj (class: measure): an object that carries the necessary values and data structures to compute 3/4 PCFs
        normalize (bool, optional): flag to determine whether we would like to normalize by the bin volume(s). Defaults to True.
        verbose_flag (bool, optional): flag to walk the user through the full calculation process. Defaults to True.
        skip_prepare (bool, optional): Flag to determine whether or not we should skip preparing the data (if you already ran the calculation). Defaults to False.

    Raises:
        AssertionError: in order to calculate the full 4PCF one has to save a boundsandnumber file (this comes from creating radial bins in one of the utility functions.)

    Returns:
        No Returns, instead we add the zeta attribute to the measure_obj.
    """    
    
    ##############################################################
    #                Projected Zeta Calculation
    ##############################################################
    if measure_obj.projected:
        start = time.time()
        if not skip_prepare:
            print("""Preparing the data:""")
            projected_prepare_data(measure_obj,verbose_flag)
        else:
            #fix bounds and number scope issue
            measure_obj.kernel_name = measure_obj.save_name
            # measure_obj.boundsandnumber = np.load(measure_obj.save_dir + 'bin_bounds_and_pixel_number_'+measure_obj.save_name+'.npy')
            
            
        if measure_obj.nPCF == 3:
            if verbose_flag:
                print("Starting Calculation of Projected 3PCF.")

                        
            PCF_computed = np.zeros((measure_obj.nbins, measure_obj.nbins, measure_obj.m_max)) + 0j
            for bin1 in range(measure_obj.nbins):
                for bin2 in range(bin1):
                    for m in range(measure_obj.m_max):
                        PCF_computed[bin1, bin2, m] += np.sum(measure_obj.density_field_data*(measure_obj.conv_M[m,bin1,:,:] * 
                                                                                            measure_obj.conv_M[m,bin2,:,:].conjugate()))
                        
            PCF_computed_ = PCF_computed.copy()           
            if normalize:
                #Normalize Full
                for bin1 in range(measure_obj.nbins):
                    for bin2 in range(bin1+1):
                        for m in range(measure_obj.m_max):
                            PCF_computed_[bin1, bin2, m] /= (measure_obj.boundsandnumber[1,bin1]*measure_obj.boundsandnumber[1,bin2])
                            PCF_computed_[bin2, bin1, m] = PCF_computed_[bin1, bin2, m]

                measure_obj.zeta = PCF_computed_
            else:
                measure_obj.zeta = PCF_computed_
                
            stop = time.time()

            print("\nFinished Calculating the Projected 3PCF in {0:0.4f} seconds".format(stop - start))
            # coeff_1d = PCF_computed_1d_
                
        elif measure_obj.nPCF == 4:
            if verbose_flag:
                print("Starting Calculation of Projected 4PCF.")

            #compute the 4PCF
            PCF_computed = np.zeros((measure_obj.nbins, measure_obj.nbins, measure_obj.nbins, 2*measure_obj.m_max, 2*measure_obj.m_max)) + 0j
            for bin1 in range(measure_obj.nbins):
                for bin2 in range(bin1):
                    for bin3 in range(bin2):
                        for m1 in range(-measure_obj.m_max,measure_obj.m_max):
                            if m1 > 0:
                                c_m1 = measure_obj.conv_M[m1, bin1, :,:]
                            else:
                                c_m1 = (-1)**m1 * measure_obj.conv_M[-m1, bin1, :,:].conjugate()

                            for m2 in range(-measure_obj.m_max,measure_obj.m_max):
                                if m2 > 0:
                                    c_m2 = measure_obj.conv_M[m2, bin2, :,:]
                                else:
                                    c_m2 = (-1)**m2 * measure_obj.conv_M[-m2, bin2, :,:].conjugate()

                                m3 = -(m1 + m2)
                                if abs(m3) >= measure_obj.m_max:
                                    continue
                                else:
                                    if m3 > 0:
                                        c_m3 = measure_obj.conv_M[m3, bin3, :,:]
                                    else:
                                        c_m3 = (-1)**m3 * measure_obj.conv_M[-m3, bin3, :,:].conjugate()

                                    PCF_computed[bin1, bin2, bin3, m1, m2] += np.sum(measure_obj.density_field_data*(c_m1*c_m2*c_m3))
             
            PCF_computed_ = PCF_computed.copy()
            if normalize:
                #Normalize and Fill
                for bin1 in range(measure_obj.nbins):
                    for bin2 in range(bin1):
                        for bin3 in range(bin2):
                            for m1 in range(-measure_obj.m_max,measure_obj.m_max):
                                for m2 in range(-measure_obj.m_max,measure_obj.m_max):
                                    m3 = -(m1 + m2)
                                    if abs(m3) >= measure_obj.m_max:
                                        continue
                                    else:
                                        PCF_computed_[bin1, bin2, bin3, m1, m2] /= (measure_obj.boundsandnumber[1,bin1]*measure_obj.boundsandnumber[1,bin2]*measure_obj.boundsandnumber[1,bin3])
                                        PCF_computed_[bin3, bin1, bin2, m1, m2] = PCF_computed_[bin1, bin2, bin3, m1, m2]
                                        PCF_computed_[bin2, bin3, bin1, m1, m2] = PCF_computed_[bin1, bin2, bin3, m1, m2]
                                        PCF_computed_[bin1, bin3, bin2, m1, m2] = PCF_computed_[bin1, bin2, bin3, m1, m2]
                                        PCF_computed_[bin2, bin1, bin3, m1, m2] = PCF_computed_[bin1, bin2, bin3, m1, m2]
                                        PCF_computed_[bin3, bin2, bin1, m1, m2] = PCF_computed_[bin1, bin2, bin3, m1, m2]


                measure_obj.zeta = PCF_computed_
            else:
                measure_obj.zeta = PCF_computed_

            stop = time.time()

            print("\nFinished Calculating the Projected 4PCF in {0:0.4f} seconds".format(stop - start))
                
            
    ##############################################################
    #                   Full Zeta Calculation
    ##############################################################        
    else:

        if not skip_prepare:
            print("""Preparing the data:""")
            prepare_data(measure_obj,verbose_flag)
        else:
            measure_obj.kernel_name = measure_obj.save_name
            # if not hasattr(measure_obj, 'boundsandnumber'):
            #     measure_obj.boundsandnumber = np.load(measure_obj.save_dir + 'bin_bounds_and_pixel_number_'+measure_obj.save_name+'.npy')


        if measure_obj.nPCF == 3:
            start = time.time()

            #load alms (results of convolution at fixed ell and for all m, to combine).
            zeta = np.zeros((measure_obj.ell_max + 1, measure_obj.nbins, measure_obj.nbins)) + 0j

            for l in range(0, measure_obj.ell_max+1, 1):
                for bin1 in range(0, measure_obj.nbins, 1):
                    for bin2 in range(0, bin1, 1):
                        for m in range(0,l+1, 1):
                            #load b1 and b2 at this m

                            ylm_b1 = np.load(measure_obj.save_dir + measure_obj.save_name +'conv_data_kernel_'+measure_obj.kernel_name+'_'+str(l)+
                                             '_'+str(m)+'_bin_'+str(bin1)+'.npy').astype(np.complex128)
                            ylm_b2 = np.load(measure_obj.save_dir + measure_obj.save_name+'conv_data_kernel_'+measure_obj.kernel_name+'_'+str(l)+
                                             '_'+str(m)+'_bin_'+str(bin2)+'.npy').astype(np.complex128)
                            #form half of sum at that m.
                            ylm_b1 *= ylm_b2.conjugate()
                            del ylm_b2

                            #average spatially now.
                            ylm_avg = np.sum(measure_obj.density_field_data*ylm_b1)
                            del ylm_b1
                            # add in negative m: so basically we are computing m's in sum term by
                            # term but with matching between + and - m.
                            if m > 0:
                                ylm_avg += ylm_avg.conjugate() 
                            zeta[l, bin1, bin2] += ylm_avg #now accumulate so we get sum over m.

            zeta *= 1./(4 * np.pi)

            for l in range(0, measure_obj.ell_max+1, 1):
                for bin1 in range(0, measure_obj.nbins, 1):
                    for bin2 in range(bin1, measure_obj.nbins, 1): #?? for bin2 in range(0, bin2<=bin1, 1):
                        zeta[l, bin1, bin2] = zeta[l, bin2, bin1]



            if normalize==True:
                binvolume = measure_obj.boundsandnumber[1,0:measure_obj.nbins]
                normedzeta = zeta / (binvolume[:,None] * binvolume[None,:])
                measure_obj.zeta = normedzeta

            else:
                measure_obj.zeta = zeta

            end = time.time()
            print("\n3PCF took {0:0.4f} seconds to finish".format(end - start))

            call('rm ' + measure_obj.save_dir + 'bin_bounds_and_pixel_number_'+measure_obj.save_name + '*', shell=True)
            call('rm ' + measure_obj.save_dir + measure_obj.save_name + 'conv_data_kernel_' + measure_obj.kernel_name + '*', shell=True)



        elif measure_obj.nPCF == 4:

            """
            This is the the big calculation for measuring the 4PCF coefficients.
            This code is adapted from Philcox et al. 2021 (Encore Paper)
            """

            def S(m):
                """
                A simple function defined in section 4.1.1 of
                Philcox et al. 2021 above algorithm 3. (Helper Function)
                """
                if m == 0:
                    return 1/2
                return 1
                

            global load_almb
            def load_almb(index_list):
                l_1, l_2, l_3, m_1, m_2, m_3, b_1, b_2, b_3 = index_list

                coupling_w = measure_obj.density_field_data * (-1)**(l_1 + l_2 + l_3) * CG_Coefficients[l_1,l_2,l_3,m_1,m_2,m_3]

                if m_1 < 0:
                    a_lmb_1 = (-1)**m_1 * (np.load(measure_obj.save_dir + measure_obj.save_name+'conv_data_kernel_'+measure_obj.kernel_name+'_'+
                                                str(l_1)+'_'+str(-m_1)+'_bin_'+str(b_1)+'.npy').astype(np.complex128)).conjugate()                                  
                else:
                    a_lmb_1 = np.load(measure_obj.save_dir + measure_obj.save_name+'conv_data_kernel_'+measure_obj.kernel_name+'_'+str(l_1)+
                                                    '_'+str(m_1)+'_bin_'+str(b_1)+'.npy').astype(np.complex128)

                if m_2 < 0:
                    a_lmb_2 = (-1)**m_2 * (np.load(measure_obj.save_dir + measure_obj.save_name+'conv_data_kernel_'+measure_obj.kernel_name+'_'+
                                                    str(l_2)+'_'+str(-m_2)+'_bin_'+str(b_2)+'.npy').astype(np.complex128)).conjugate()                               
                else:
                    a_lmb_2 = np.load(measure_obj.save_dir + measure_obj.save_name+'conv_data_kernel_'+measure_obj.kernel_name+'_'+str(l_2)+
                                                        '_'+str(m_2)+'_bin_'+str(b_2)+'.npy').astype(np.complex128)

                a_lmb_3 = np.load(measure_obj.save_dir + measure_obj.save_name+'conv_data_kernel_'+measure_obj.kernel_name+'_'+str(l_3)+
                                                        '_'+str(m_3)+'_bin_'+str(b_3)+'.npy').astype(np.complex128)

                # zeta[l_1, l_2, l_3, b_1, b_2, b_3] += np.sum(2 * S(m_3) * coupling_w * np.real(a_lmb_1 * a_lmb_2 * a_lmb_3))
                return [[l_1, l_2, l_3, b_1, b_2, b_3], np.sum(2 * S(m_3) * coupling_w * np.real(a_lmb_1 * a_lmb_2 * a_lmb_3))]

            

            #we define these local variables out of laziness to change my original non-object oriented code
            ell_max = measure_obj.ell_max
            nbins = measure_obj.nbins
            if hasattr(measure_obj, 'boundsandnumber') == False:
                raise AssertionError("You need to run measure_obj.create_radial_bins() first!")

            binvolume = measure_obj.boundsandnumber[1,0:nbins]
            zeta = np.zeros((ell_max+1, ell_max+1, ell_max+1,nbins, nbins, nbins)) + 0j
            stream = pkg_resources.resource_stream(__name__, 'utils/CG_Coeffs.npy')
            CG_Coefficients = np.load(stream)

            indeces = []
            for l_1 in range(0,ell_max+1):
                for l_2 in range(0,ell_max+1):
                    for l_3 in range(np.abs(l_1 - l_2), min(l_1 + l_2, ell_max)+1):
                        if (l_1 + l_2 + l_3)%2 != 0: # we don't assume this to be true for Turbulent ISM
                            continue
                        for m_1 in range(-l_1, l_1 + 1):
                            for m_2 in range(-l_2, l_2 + 1):
                                m_3 = -m_1 - m_2
                                if m_3 > l_3 or m_3 < 0:
                                    continue
                                for b_1 in range(0, nbins):        
                                    for b_2 in range(b_1+1, nbins):
                                        for b_3 in range(b_2+1, nbins):
                                            indeces.append((l_1, l_2, l_3, 
                                                            m_1, m_2, m_3,
                                                            b_1, b_2, b_3))



            start = time.time()
            print("Executing 4PCF Calculation ...")

            with concurrent.futures.ProcessPoolExecutor() as executor:
                results = list(executor.map(load_almb, indeces))
            
            executor.shutdown(wait=True)

            for j in range(len(results)):
                l_1, l_2, l_3, b_1, b_2, b_3 = results[j][0]
                zeta[l_1, l_2, l_3, b_1, b_2, b_3] += results[j][1] 


            for l1 in range(0,ell_max+1):
                for l2 in range(0,ell_max+1):
                    for l3 in range(np.abs(l1 - l2), min(l1 + l2, ell_max)+1):
                        if (l1 + l2 + l3)%2 != 0: # we don't assume this to be true for Turbulent ISM
                            continue
                        for b1 in range(0,nbins):
                            for b2 in range(b1+1,nbins):
                                for b3 in range(b2+1,nbins):
                                    this_4pcf = zeta[l1,l2,l3,b1,b2,b3]
                                    zeta[l3,l1,l2,b3,b1,b2] = this_4pcf
                                    zeta[l2,l3,l1,b2,b3,b1] = this_4pcf
                                    zeta[l1,l3,l2,b1,b3,b2] = this_4pcf
                                    zeta[l2,l1,l3,b2,b1,b3] = this_4pcf
                                    zeta[l3,l2,l1,b3,b2,b1] = this_4pcf

            finish=time.time()

            call('rm ' + measure_obj.save_dir + 'bin_bounds_and_pixel_number_'+measure_obj.save_name + '*', shell=True)
            call('rm ' + measure_obj.save_dir + measure_obj.save_name + 'conv_data_kernel_' + measure_obj.kernel_name + '*', shell=True)

            print("Finished Calculating 4PCF in {0:0.4f} seconds".format(finish-start))
            if normalize:
                """
                Normalize zeta^L_B (where L = {\ell_1, \ell_2, \ell_3} and B = {b_1, b_2, b_3}) hat
                coefficients from calc_zeta by dividing by bin volume
                """
                #binvolume = measure_obj.boundsandnumber[1,0:nbins]
                #normalize_coeff = (binvolume[:,None, None] * binvolume[None,:, None] * binvolume[None, None, :])
                normalize_coeff = (4.*np.pi)**(3.)
                zeta = (normalize_coeff*zeta/((measure_obj.ld_one_d**3)))
                measure_obj.zeta = zeta
            else: 
                print("your zeta coefficients are not properly ormalized.")
                measure_obj.zeta = zeta    


            
            
            
            



import numpy as np
from sarabande.utils.functions import *

class measure:

    
    def __init__(self, nPCF=4, projected=False, m_max=None, density_field_data = None, 
                 save_dir=None, save_name=None, ell_max=5, nbins=4, bin_spacing='LIN',
                 bin_min=1, physical_boxsize = None, rmin = None, rmax = None):

        """
        This class allows us to measure the 3/4pcf from some input data field

        Args:
            nPCF ([int]): Must be either 3 or 4. Determines how many points we use in our nPCF.
            projected ([bool]): Flag to determine whether the user wants a projected 3/4 PCF or the Full. Defaults to False.
                if projected:
                    m_max ([int]): If user chooses projected, we set an m_max (similar to the ell_max in 3D)
                if not projected:
                    ell_max ([int]): If user choosees not projected (full nPCF) then ell_max is the highest order for calculation.
            density_field_data ([ndarray]): A square ndarray of data that is periodic. Must be 2D for projected and 3D for full.
            save_dir ([string]): A string to tell the algorithm where to save and store files. All temporary files will be stored here.
            save_name ([string]): A string to tell the algorithm what to name the files.
            nbins ([int]): Number of bins to be used in nPCF calculation.
            bin_spacing ([string]): A string to determine the spacing of bins. Options are 'LIN', 'INV', or 'LOG'
            bin_min ([int]): The lower bound of the inner most bin. Default is 1. Optional.
            physical_boxsize ([float]): An optional parameter if using a physical scale. The length of one side of the data.
            rmin ([float]): minimum calculation distance (determins bin_min)
            rmax ([float]): maximum calculation distance (determins bin_max)
     
        """                 

        self.ell_max = ell_max
        self.eps = 1e-15
        self.nbins = nbins
        self.projected = projected
        self.ld_one_d = np.shape(density_field_data)[0]
        self.bin_min = bin_min-1e-5
        self.bin_max = (self.ld_one_d // 2) + 1e-5
        
        ####################################
        #   Initialization Case Handling
        ####################################
        
        if nPCF == 3 or nPCF == 4:
            self.nPCF = nPCF
            if self.projected:
                if m_max is None:
                    raise ValueError("You need to provide an m_max you would like to compute up to.")
                else:
                    self.m_max = m_max
                    
        else:
            raise ValueError("Sarabande only calculates 3 or 4 point correlation functions. Please give an integer 3 or 4.")

        
        if physical_boxsize or rmin or rmax is not None:
            if physical_boxsize and rmin and rmax is not None:
                self.bin_min = (rmin/physical_boxsize)*self.ld_one_d - 1e-5
                self.bin_max = (rmax/physical_boxsize)*self.ld_one_d + 1e-5  
            else:
                raise AssertionError("""If you want to use physical scales, you need to give physical_boxsize, rmin, and rmax""")
        
        if bin_spacing == 'LIN' or bin_spacing == 'INV' or bin_spacing == 'LOG':
            #We can toggle what binning we want to use using the bin_spacing argument
            switch = {
            'LIN' : np.linspace(self.bin_min, self.bin_max, self.nbins+1),
            'INV' : 1./np.linspace(1./self.bin_min, 1./self.bin_max, self.nbins+1),
            'LOG' : np.exp(np.linspace(np.log(self.bin_min), np.log(self.bin_max), self.nbins+1))}
        else:
            raise ValueError("""Please put a valid bin_spacing argument, acceptable options are: \n LIN \n INV \n LOG \n in string format.""")
        
        self.bin_edges = switch[bin_spacing]
        # self.ld_one_d = ld_one_d
        
        if density_field_data is not None:
            if len(np.shape(density_field_data)) == 3 and self.projected == True:
                raise AssertionError("""Projected 3/4 PCFs can only be computed on a 2D data set, use full 3/4 PCFs for 3D data sets.""")
            elif len(np.shape(density_field_data)) == 2 and self.projected == False:
                raise AssertionError("""Projected 3/4 PCFs can only be computed on a 2D data set, use full 3/4 PCFs for 3D data sets.""")
            else:
                self.density_field_data = density_field_data
            
        else:
            if self.projected == True:
                raise ValueError("Please include a density_field_data argument. Should be a density sheet in the form of a numpy array")
            else:
                raise ValueError("Please include a density_field_data argument. Should be a density cube in the form of a numpy array")
        
        if save_name is not None:
            self.save_name = save_name
        else:
            raise ValueError("Please include a save_name argument")
            
        if save_dir is not None:
            self.save_dir = save_dir
        else:
            raise ValueError("Please include a save_dir argument")
            
    
    


# """
# -----------
# Sarabande - 0.0.5
# --------------------------------------------
# Developers: J.Sunseri
# Contact: James Sunseri (jamessunseri@berkeley.edu)
# License: MIT
# ---------------------------------------------
# Tool for measuring 3/4 PCFs on discrete periodic data..
# ---------------------------------------------

# """

# # import numpy as np
# # from subprocess import call
# # from .utils import *
# # from main import measure
# # from full_PCF import calc_zeta
# # from full_PCF_mapped import calc_zeta_mapped
# # from full_PCF_parallel import calc_zeta_parallel

# from . import full_PCF_mapped
# from . import full_PCF_parallel
# from . import full_PCF

# def about():
#     text = "-------------------------------\n"
#     text += "-    Welcome to Sarabande!    -\n"
#     text += "-------------------------------\n"
#     text += "Download the latest version from: https://pypi.org/project/sarabande/\n"
#     text += "Report issues to: https://github.com/James11222/sarabande\n"
#     text += "Read the documentations at: " + "https://github.com/James11222/sarabande" + "\n"
#     text += "-------------------------------\n"
#     print(text)


# if __name__ == "__main__":
#     about()
